"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.integration import BaseCliTestCase

from cfnlint import ConfigMixIn


class TestQuickStartTemplates(BaseCliTestCase):
    """Test QuickStart Templates Parsing"""

    scenarios = [
        {
            "filename": "test/fixtures/templates/public/lambda-poller.yaml",
            "results_filename": "test/fixtures/results/public/lambda-poller.json",
            "exit_code": 2,
        },
        {
            "filename": "test/fixtures/templates/public/watchmaker.json",
            "results_filename": "test/fixtures/results/public/watchmaker.json",
            "exit_code": 8,
        },
        {
            "filename": "test/fixtures/templates/quickstart/nist_high_main.yaml",
            "results_filename": "test/fixtures/results/quickstart/nist_high_main.json",
            "exit_code": 14,
        },
        {
            "filename": "test/fixtures/templates/quickstart/nist_application.yaml",
            "results_filename": (
                "test/fixtures/results/quickstart" "/nist_application.json"
            ),
            "exit_code": 14,
        },
        {
            "filename": "test/fixtures/templates/quickstart/nist_config_rules.yaml",
            "results_filename": (
                "test/fixtures/results/quickstart/" "nist_config_rules.json"
            ),
            "exit_code": 6,
        },
        {
            "filename": "test/fixtures/templates/quickstart/nist_iam.yaml",
            "results_filename": "test/fixtures/results/quickstart/nist_iam.json",
            "exit_code": 4,
        },
        {
            "filename": "test/fixtures/templates/quickstart/nist_logging.yaml",
            "results_filename": "test/fixtures/results/quickstart/nist_logging.json",
            "exit_code": 14,
        },
        {
            "filename": "test/fixtures/templates/quickstart/nist_vpc_management.yaml",
            "results_filename": (
                "test/fixtures/results/quickstart/" "nist_vpc_management.json"
            ),
            "exit_code": 14,
        },
        {
            "filename": "test/fixtures/templates/quickstart/nist_vpc_production.yaml",
            "results_filename": (
                "test/fixtures/results/quickstart/" "nist_vpc_production.json"
            ),
            "exit_code": 14,
        },
        {
            "filename": "test/fixtures/templates/quickstart/openshift_master.yaml",
            "results_filename": (
                "test/fixtures/results/quickstart" "/openshift_master.json"
            ),
            "exit_code": 8,
        },
        {
            "filename": "test/fixtures/templates/quickstart/openshift.yaml",
            "results_filename": "test/fixtures/results/quickstart/openshift.json",
            "exit_code": 14,
        },
        {
            "filename": "test/fixtures/templates/quickstart/cis_benchmark.yaml",
            "results_filename": "test/fixtures/results/quickstart/cis_benchmark.json",
            "exit_code": 6,
        },
    ]

    def test_templates(self):
        """Test same templates using integration approach"""
        self.run_module_integration_scenarios(
            ConfigMixIn(
                [],
                include_checks=["I"],
                configure_rules={"E3012": {"strict": True}},
                include_experimental=True,
            )
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any, Iterator

from cfnlint.jsonschema import ValidationError, ValidationResult
from cfnlint.jsonschema.protocols import Validator
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class TaskDefinitionAwsVpc(CfnLintKeyword):
    id = "E3053"
    shortdesc = "Validate ECS task definition is has correct values for 'HostPort'"
    description = (
        "The 'HostPort' must either be undefined or equal to "
        "the 'ContainerPort' value"
    )
    tags = ["resources", "ecs"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::ECS::TaskDefinition/Properties"],
        )

    def _get_port_mappings(
        self, validator: Validator, instance: Any
    ) -> Iterator[tuple[str | int | None, str | int | None, Validator]]:

        for container_definition, container_definition_validator in get_value_from_path(
            validator,
            instance,
            path=deque(["ContainerDefinitions", "*", "PortMappings", "*"]),
        ):
            for host_port, host_port_validator in get_value_from_path(
                container_definition_validator,
                container_definition,
                path=deque(["HostPort"]),
            ):
                if not isinstance(host_port, (str, int)):
                    continue
                for container_port, _ in get_value_from_path(
                    host_port_validator,
                    container_definition,
                    path=deque(["ContainerPort"]),
                ):
                    if not isinstance(container_port, (str, int)):
                        continue
                    if str(host_port) != str(container_port):
                        yield host_port, container_port, host_port_validator

    def validate(
        self, validator: Validator, _: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        for network_mode, _ in get_value_from_path(
            validator,
            instance,
            path=deque(["NetworkMode"]),
        ):
            if network_mode != "awsvpc":
                continue
            for (
                host_port,
                container_port,
                port_mapping_validator,
            ) in self._get_port_mappings(
                validator,
                instance,
            ):
                yield ValidationError(
                    f"{host_port!r} does not equal {container_port!r}",
                    validator="const",
                    rule=self,
                    path_override=port_mapping_validator.context.path.path,
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class Types(CloudFormationLintRule):
    """Check if Parameters are typed"""

    id = "E2002"
    shortdesc = "Parameters have appropriate type"
    description = "Making sure the parameters have a correct type"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/best-practices.html#parmtypes"
    tags = ["parameters"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint import ConfigMixIn
from cfnlint.rules.functions.SubNeeded import SubNeeded  # pylint: disable=E0401


class TestSubNeeded(BaseRuleTestCase):
    """Test Rules Get Att"""

    def setUp(self):
        """Setup"""
        super(TestSubNeeded, self).setUp()
        self.collection.register(SubNeeded())
        self.success_templates = [
            "test/fixtures/templates/good/functions/sub.yaml",
            "test/fixtures/templates/good/functions/sub_needed.yaml",
            "test/fixtures/templates/good/functions/sub_needed_transform.yaml",
        ]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_template_config(self):
        """Test custom excludes configuration"""
        self.helper_file_positive_template(
            "test/fixtures/templates/good/functions/sub_needed_custom_excludes.yaml",
            ConfigMixIn(
                [],
                configure_rules={
                    "E1029": {"custom_excludes": "^\\$\\{Stage\\}$"},
                },
            ),
        )
        self.helper_file_positive_template(
            "test/fixtures/templates/good/functions/sub_needed_custom_excludes.yaml",
            ConfigMixIn(
                [],
                configure_rules={
                    "E1029": {"custom_excludes": "^\\$\\{Stage\\}$"},
                },
            ),
        )
        self.helper_file_negative(
            "test/fixtures/templates/good/functions/sub_needed_custom_excludes.yaml",
            1,
            ConfigMixIn([]),
        )

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/functions/sub_needed.yaml", 7
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.Modules import Modules


class TestModules(BaseRuleTestCase):
    def setUp(self):
        """Setup"""
        super(TestModules, self).setUp()
        self.collection.register(Modules())
        self.success_templates = [
            "test/fixtures/templates/good/modules/minimal.yaml",
        ]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_cannot_have_update_policy(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/modules/bad_has_update_policy.yaml", 1
        )

    def test_cannot_have_create_policy(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/modules/bad_has_create_policy.yaml", 1
        )

    def test_cannot_use_special_metadata_keyword(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/modules/bad_uses_module_metadata.yaml", 1
        )

    def test_cannot_use_tags(self):
        """Test failure"""
        self.helper_file_negative(
            "test/fixtures/templates/bad/modules/bad_has_tags.yaml", 1
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.rds.DbClusterInstanceClassEnum import (
    DbClusterInstanceClassEnum,
)


@pytest.fixture(scope="module")
def rule():
    rule = DbClusterInstanceClassEnum()
    rule._schema = {
        "us-east-1": {
            "allOf": [
                {
                    "if": {
                        "properties": {
                            "DBClusterInstanceClass": {"type": "string"},
                            "Engine": {"const": "mysql"},
                        },
                        "required": ["Engine", "DBClusterInstanceClass"],
                    },
                    "then": {
                        "properties": {
                            "DBClusterInstanceClass": {
                                "enum": [
                                    "db.m5d.12xlarge",
                                    "db.m5d.16xlarge",
                                ]
                            }
                        }
                    },
                }
            ]
        }
    }
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {"Engine": "mysql", "DBClusterInstanceClass": "db.m5d.12xlarge"},
            [],
        ),
        (
            {"Engine": "aurora-mysql", "DBClusterInstanceClass": "a"},
            [],
        ),
        (
            {
                "Engine": {"Ref": "Engine"},
                "DBClusterInstanceClass": {"Ref": "DBInstanceClass"},
            },
            [],
        ),
        (
            {
                "Engine": "mysql",
                "DBClusterInstanceClass": {"Ref": "DBInstanceClass"},
            },
            [],
        ),
        (
            {
                "Engine": "mysql",
                "DBClusterInstanceClass": "a",
            },
            [
                ValidationError(
                    (
                        "'a' is not one of ['db.m5d.12xlarge', "
                        "'db.m5d.16xlarge'] in 'us-east-1'"
                    ),
                    rule=DbClusterInstanceClassEnum(),
                    path=deque(["DBClusterInstanceClass"]),
                    validator="enum",
                    schema_path=deque(
                        [
                            "allOf",
                            0,
                            "then",
                            "properties",
                            "DBClusterInstanceClass",
                            "enum",
                        ]
                    ),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import subprocess
from test.integration import BaseCliTestCase


class TestFormatters(BaseCliTestCase):
    """Test Formatters"""

    def test_junit(self):
        """Test JUnit formatting"""
        result = subprocess.check_output(
            [
                "cfn-lint",
                "--format",
                "junit",
                "--",
                "test/fixtures/templates/good/core/config_only_*.yaml",
            ]
        )

        if isinstance(result, bytes):
            result = result.decode("utf8")

        self.assertIn(
            (
                '<testcase name="I1002 Validate approaching the template size limit"'
                ' url="https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"/>'
            ),
            result,
        )
        self.assertIn(
            (
                '<testcase name="I1003 Validate if we are approaching the '
                'max size of a description" '
                'url="https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"/>'
            ),
            result,
        )

    def test_pretty(self):
        """Test pretty formatting"""
        result = subprocess.check_output(
            [
                "cfn-lint",
                "--format",
                "pretty",
                "--",
                "test/fixtures/templates/good/core/config_only_*.yaml",
            ]
        )

        if isinstance(result, bytes):
            result = result.decode("utf8")

        self.assertEqual(
            (
                "Cfn-lint scanned 2 templates against 2 rules and found 0 errors, 0"
                " warnings, and 0 informational violations"
            ),
            result.strip(),
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_ecs_taskdefinition
from cfnlint.jsonschema import ValidationError
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class FargateCpuMemory(CfnLintJsonSchema):
    id = "E3047"
    shortdesc = (
        "Validate ECS Fargate tasks have the right combination of CPU and memory"
    )
    description = (
        "When using a ECS Fargate task there is a specfic combination "
        "of memory and cpu that can be used"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ecs-taskdefinition.html#cfn-ecs-taskdefinition-memory"
    tags = ["properties", "ecs", "service", "container", "fargate"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::ECS::TaskDefinition/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_ecs_taskdefinition,
                filename="fargate_cpu_memory.json",
            ),
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return (
            f"Cpu {instance.get('Cpu')!r} is not "
            "compatible with memory "
            f"{instance.get('Memory')!r}"
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules._rule import CloudFormationLintRule


class ParseError(CloudFormationLintRule):
    """Parse Lint Rule"""

    id = "E0000"
    shortdesc = "Parsing error found when parsing the template"
    description = "Checks for JSON/YAML formatting errors in your template"
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["base"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging

import pytest

from cfnlint.schema._schema import Schema

LOGGER = logging.getLogger("cfnlint.schema.manager")
LOGGER.disabled = True


@pytest.fixture
def resource_schema():
    return {
        "additionalProperties": False,
        "definitions": {
            "ListItem": {"description": "A a list Id.", "type": "string"},
            "ResourceArn": {
                "description": "A resource ARN.",
                "maxLength": 256,
                "minLength": 1,
                "pattern": "^arn:aws.*$",
                "type": "string",
            },
        },
        "properties": {
            "Name": {
                "maxLength": 128,
                "minLength": 1,
                "pattern": "^[a-zA-Z0-9-]+$",
                "type": "string",
            },
            "List": {
                "insertionOrder": False,
                "items": {"$ref": "#/definitions/ListItem"},
                "type": "array",
            },
            "Arn": {"$ref": "#/definitions/ResourceArn"},
            "Id": {
                "maxLength": 36,
                "minLength": 36,
                "pattern": "^([0-9a-f]{8})-([0-9a-f]{4}-){3}([0-9a-f]{12})$",
                "type": "string",
            },
        },
        "readOnlyProperties": [
            "/properties/Arn",
            "/properties/Id",
            "/properties/List",
        ],
        "typeName": "AWS::NetworkFirewall::Firewall",
    }


def test_schema(resource_schema):

    schema = Schema(schema=resource_schema)
    assert list(schema.get_atts.keys()) == ["Arn", "Id", "List"]
    assert schema.get_atts["Arn"] == "/properties/Arn"
    assert schema.get_atts["Id"] == "/properties/Id"
    assert schema.get_atts["List"] == "/properties/List"


@pytest.fixture
def resource_vpc_schema():
    return {
        "additionalProperties": False,
        "conditionalCreateOnlyProperties": ["/properties/InstanceTenancy"],
        "createOnlyProperties": [
            "/properties/CidrBlock",
            "/properties/Ipv4IpamPoolId",
            "/properties/Ipv4NetmaskLength",
        ],
        "definitions": {
            "Tag": {
                "additionalProperties": False,
                "properties": {"Key": {"type": "string"}, "Value": {"type": "string"}},
                "required": ["Value", "Key"],
                "type": "object",
            }
        },
        "dependentRequired": {"Ipv4IpamPoolId": ["Ipv4NetmaskLength"]},
        "primaryIdentifier": ["/properties/VpcId"],
        "properties": {
            "CidrBlock": {"type": "string"},
            "CidrBlockAssociations": {
                "insertionOrder": False,
                "items": {"type": "string"},
                "type": "array",
                "uniqueItems": False,
            },
            "DefaultNetworkAcl": {"insertionOrder": False, "type": "string"},
            "DefaultSecurityGroup": {
                "format": "AWS::EC2::SecurityGroup.GroupId",
                "insertionOrder": False,
                "type": "string",
            },
            "EnableDnsHostnames": {"type": "boolean"},
            "EnableDnsSupport": {"type": "boolean"},
            "InstanceTenancy": {"type": "string"},
            "Ipv4IpamPoolId": {"type": "string"},
            "Ipv4NetmaskLength": {"type": "integer"},
            "Ipv6CidrBlocks": {
                "insertionOrder": False,
                "items": {"type": "string"},
                "type": "array",
                "uniqueItems": False,
            },
            "Tags": {
                "insertionOrder": False,
                "items": {"$ref": "#/definitions/Tag"},
                "type": "array",
                "uniqueItems": False,
            },
            "VpcId": {"format": "AWS::EC2::VPC.Id", "type": "string"},
        },
        "readOnlyProperties": [
            "/properties/CidrBlockAssociations",
            "/properties/DefaultNetworkAcl",
            "/properties/DefaultSecurityGroup",
            "/properties/Ipv6CidrBlocks",
            "/properties/VpcId",
        ],
        "requiredXor": ["CidrBlock", "Ipv4IpamPoolId"],
        "tagging": {
            "cloudFormationSystemTags": True,
            "tagOnCreate": True,
            "tagProperty": "/properties/Tags",
            "tagUpdatable": True,
            "taggable": True,
        },
        "typeName": "AWS::EC2::VPC",
        "writeOnlyProperties": [
            "/properties/Ipv4IpamPoolId",
            "/properties/Ipv4NetmaskLength",
        ],
    }


def test_vpc_schema(resource_vpc_schema):
    schema = Schema(schema=resource_vpc_schema)
    assert list(schema.get_atts.keys()) == [
        "CidrBlock",
        "CidrBlockAssociations",
        "DefaultNetworkAcl",
        "DefaultSecurityGroup",
        "EnableDnsHostnames",
        "EnableDnsSupport",
        "InstanceTenancy",
        "Ipv4IpamPoolId",
        "Ipv4NetmaskLength",
        "Ipv6CidrBlocks",
        "Tags",
        "VpcId",
    ]
    assert schema.get_atts["VpcId"] == "/properties/VpcId"
    assert schema.get_atts["CidrBlock"] == "/properties/CidrBlock"
    assert schema.get_atts["Ipv6CidrBlocks"] == "/properties/Ipv6CidrBlocks"


@pytest.fixture
def resource_servicecatalog_schema():
    return {
        "additionalProperties": False,
        "createOnlyProperties": [
            "/properties/NotificationArns",
            "/properties/ProvisionedProductName",
        ],
        "definitions": {
            "OutputType": {"type": "string"},
            "ProvisioningParameter": {
                "additionalProperties": False,
                "properties": {
                    "Key": {"maxLength": 1000, "minLength": 1, "type": "string"},
                    "Value": {"maxLength": 4096, "type": "string"},
                },
                "required": ["Key", "Value"],
                "type": "object",
            },
            "ProvisioningPreferences": {
                "additionalProperties": False,
                "properties": {
                    "StackSetAccounts": {
                        "items": {"pattern": "^[0-9]{12}$", "type": "string"},
                        "type": "array",
                        "uniqueItems": True,
                    },
                    "StackSetFailureToleranceCount": {"minimum": 0, "type": "integer"},
                    "StackSetFailureTolerancePercentage": {
                        "maximum": 100,
                        "minimum": 0,
                        "type": "integer",
                    },
                    "StackSetMaxConcurrencyCount": {"minimum": 1, "type": "integer"},
                    "StackSetMaxConcurrencyPercentage": {
                        "maximum": 100,
                        "minimum": 1,
                        "type": "integer",
                    },
                    "StackSetOperationType": {
                        "enum": ["CREATE", "UPDATE", "DELETE"],
                        "type": "string",
                    },
                    "StackSetRegions": {
                        "items": {
                            "pattern": "^[a-z]{2}-([a-z]+-)+[1-9]",
                            "type": "string",
                        },
                        "type": "array",
                        "uniqueItems": True,
                    },
                },
                "type": "object",
            },
            "Tag": {
                "additionalProperties": False,
                "properties": {
                    "Key": {
                        "maxLength": 128,
                        "minLength": 1,
                        "pattern": "^([\\p{L}\\p{Z}\\p{N}_.:/=+\\-@]*)$",
                        "type": "string",
                    },
                    "Value": {
                        "maxLength": 256,
                        "minLength": 1,
                        "pattern": "^([\\p{L}\\p{Z}\\p{N}_.:/=+\\-@]*)$",
                        "type": "string",
                    },
                },
                "required": ["Key", "Value"],
                "type": "object",
            },
        },
        "documentationUrl": "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-servicecatalog-cloudformationprovisionedproduct.html",
        "handlers": {
            "create": {"permissions": ["*"], "timeoutInMinutes": 720},
            "delete": {"permissions": ["*"]},
            "read": {"permissions": ["*"]},
            "update": {"permissions": ["*"], "timeoutInMinutes": 720},
        },
        "primaryIdentifier": ["/properties/ProvisionedProductId"],
        "properties": {
            "AcceptLanguage": {"enum": ["en", "jp", "zh"], "type": "string"},
            "CloudformationStackArn": {
                "maxLength": 256,
                "minLength": 1,
                "type": "string",
            },
            "NotificationArns": {
                "items": {"type": "string"},
                "maxItems": 5,
                "type": "array",
                "uniqueItems": True,
            },
            "Outputs": {
                "additionalProperties": False,
                "maxProperties": 100,
                "patternProperties": {
                    "^[A-Za-z0-9]{1,64}$": {"$ref": "#/definitions/OutputType"}
                },
                "type": "object",
            },
            "PathId": {"maxLength": 100, "minLength": 1, "type": "string"},
            "PathName": {"maxLength": 100, "minLength": 1, "type": "string"},
            "ProductId": {"maxLength": 100, "minLength": 1, "type": "string"},
            "ProductName": {"maxLength": 128, "minLength": 1, "type": "string"},
            "ProvisionedProductId": {"maxLength": 50, "minLength": 1, "type": "string"},
            "ProvisionedProductName": {
                "maxLength": 128,
                "minLength": 1,
                "type": "string",
            },
            "ProvisioningArtifactId": {
                "maxLength": 100,
                "minLength": 1,
                "type": "string",
            },
            "ProvisioningArtifactName": {"type": "string"},
            "ProvisioningParameters": {
                "items": {"$ref": "#/definitions/ProvisioningParameter"},
                "type": "array",
            },
            "ProvisioningPreferences": {
                "$ref": "#/definitions/ProvisioningPreferences"
            },
            "RecordId": {"maxLength": 50, "minLength": 1, "type": "string"},
            "Tags": {"items": {"$ref": "#/definitions/Tag"}, "type": "array"},
        },
        "readOnlyProperties": [
            "/properties/RecordId",
            "/properties/CloudformationStackArn",
            "/properties/Outputs",
            "/properties/ProvisionedProductId",
        ],
        "sourceUrl": "https://github.com/aws-cloudformation/aws-cloudformation-rpdk.git",
        "typeName": "AWS::ServiceCatalog::CloudFormationProvisionedProduct",
    }


def test_servicecatalog_cloudformation_schema(resource_servicecatalog_schema):

    schema = Schema(schema=resource_servicecatalog_schema)
    assert list(schema.get_atts.keys()) == [
        "RecordId",
        "CloudformationStackArn",
        "Outputs\\..*",
        "ProvisionedProductId",
    ]
    assert schema.get_atts["Outputs.Example"] == "/properties/CfnLintStringType"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import namedtuple
from typing import Any, Sequence

from cfnlint.helpers import load_resource
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.jsonschema.exceptions import best_match
from cfnlint.rules.jsonschema.Base import BaseJsonSchema

SchemaDetails = namedtuple("SchemaDetails", ["module", "filename"])


class CfnLintJsonSchema(BaseJsonSchema):
    def __init__(
        self,
        keywords: Sequence[str] | None = None,
        schema_details: SchemaDetails | None = None,
        all_matches: bool = False,
    ) -> None:
        super().__init__()
        self.keywords = keywords or []
        self.parent_rules = ["E1101"]
        self.all_matches = all_matches
        self._use_schema_arg = True
        self._schema: Any = {}

        if schema_details:
            self._schema = load_resource(
                schema_details.module,
                filename=schema_details.filename,
            )
            self._use_schema_arg = False

    @property
    def schema(self):
        return self._schema

    def message(self, instance: Any, err: ValidationError) -> str:
        return self.shortdesc

    def _iter_errors(self, validator, instance):
        errs = list(validator.iter_errors(instance))
        if not self.all_matches:
            err = best_match(errs)
            if err is not None:
                err.message = self.message(instance, err)
                err.context = []
                err.rule = self
                yield err
                return

        for err in errs:
            yield self._clean_error(err)

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        # if the schema has a description will only replace the message with that
        # description and use the best error for the location information
        if not self._use_schema_arg:
            schema = self._schema

        cfn_validator = self.extend_validator(
            validator=validator.evolve(
                function_filter=validator.function_filter.evolve(
                    validate_dynamic_references=False,
                    add_cfn_lint_keyword=False,
                )
            ),
            schema=schema,
            context=validator.context.evolve(
                functions=[],
                strict_types=True,
            ),
        )

        yield from self._iter_errors(cfn_validator, instance)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging
from test.testlib.testcase import BaseTestCase

import cfnlint.config  # pylint: disable=E0401

LOGGER = logging.getLogger("cfnlint")


class TestLogging(BaseTestCase):
    """Test Logging Arguments"""

    def tearDown(self):
        """Setup"""
        for handler in LOGGER.handlers:
            LOGGER.removeHandler(handler)

    def test_logging_info(self):
        """Test success run"""

        cfnlint.config.configure_logging(False, True)
        self.assertEqual(logging.INFO, LOGGER.level)
        self.assertEqual(len(LOGGER.handlers), 1)

    def test_logging_debug(self):
        """Test debug level"""

        cfnlint.config.configure_logging(True, False)
        self.assertEqual(logging.DEBUG, LOGGER.level)
        self.assertEqual(len(LOGGER.handlers), 1)

    def test_no_logging(self):
        """Test no logging level"""

        cfnlint.config.configure_logging(False, False)
        self.assertEqual(logging.WARNING, LOGGER.level)
        self.assertEqual(len(LOGGER.handlers), 1)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import fnmatch
import json
import os
from unittest import TestCase

import pytest

import cfnlint
from cfnlint.helpers import format_json_string


@pytest.mark.data
class TestDataFileFormat(TestCase):
    """Test format of data files"""

    def setUp(self) -> None:
        self.modules = [
            "AdditionalSpecs",
            "CfnLintCli",
            "Serverless",
            "schemas",
        ]
        super().setUp()

    def test_data_module_specs(self):
        """Test data file formats"""

        for module in self.modules:
            append_dir = os.path.join(
                os.path.dirname(cfnlint.__file__),
                "data",
                module,
            )
            for dirpath, _, filenames in os.walk(append_dir):
                for filename in fnmatch.filter(filenames, "*.json"):
                    string_content = ""
                    with open(
                        os.path.join(dirpath, filename), encoding="utf8"
                    ) as input_file:
                        string_content = "".join(input_file.readlines())

                    pretty_content = format_json_string(json.loads(string_content))
                    self.assertEqual(
                        string_content, f"{pretty_content}\n", f"{dirpath}/{filename}"
                    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import datetime

import pytest

from cfnlint.rules.resources.properties.StringLength import StringLength


@pytest.fixture(scope="module")
def rule():
    yield StringLength()


@pytest.mark.parametrize(
    "instance,mL,schema,expected",
    [
        ("foo", 2, {}, 0),
        ("foo", 4, {}, 1),
        ({"Fn::Sub": "foo"}, 2, {}, 0),
        ({"Fn::Sub": "foo"}, 4, {}, 1),
        ({"Fn::Sub": ["foo", {}]}, 2, {}, 0),
        ({"Fn::Sub": ["foo", {}]}, 4, {}, 1),
        ({"Fn::Sub": ["foo", {}, {}]}, 2, {}, 0),
        ({"foo": "bar"}, 10, {"type": "object"}, 0),
        ({"foo": "bar"}, 20, {"type": "object"}, 1),
        ({"foo": datetime.datetime.now()}, 10, {"type": "object"}, 0),
        ({"foo": datetime.datetime.now()}, 40, {"type": "object"}, 1),
        ({"foo": ["bar"]}, 10, {"type": "object"}, 0),
        ({"foo": ["bar"]}, 20, {"type": "object"}, 1),
        ({"foo": 1, "bar": {"Ref": "Parameter"}}, 10, {"type": "object"}, 0),
        ({"foo": 1, "bar": {"Ref": "Parameter"}}, 20, {"type": "object"}, 1),
        ({"foo": 1, "bar": {"Fn::Sub": "2"}}, 10, {"type": "object"}, 0),
        ({"foo": 1, "bar": {"Fn::Sub": "2"}}, 20, {"type": "object"}, 1),
        ({"foo": 1, "bar": {"Fn::Sub": ["2", {}]}}, 10, {"type": "object"}, 0),
        ({"foo": 1, "bar": {"Fn::Sub": ["2", {}]}}, 20, {"type": "object"}, 1),
    ],
)
def test_min_length(instance, mL, expected, rule, schema, validator):
    assert len(list(rule.minLength(validator, mL, instance, schema))) == expected


@pytest.mark.parametrize(
    "instance,mL,schema,expected",
    [
        ("foo", 4, {}, 0),
        ("foo", 2, {}, 1),
        ({"Fn::Sub": "foo"}, 4, {}, 0),
        ({"Fn::Sub": "foo"}, 2, {}, 1),
        ({"Fn::Sub": ["foo", {}]}, 4, {}, 0),
        ({"Fn::Sub": ["foo", {}]}, 2, {}, 1),
        ({"Fn::Sub": ["foo", {}, {}]}, 2, {}, 0),
        ({"foo": "bar"}, 20, {"type": "object"}, 0),
        ({"foo": "bar"}, 2, {"type": "object"}, 1),
        ({"foo": datetime.datetime.now()}, 40, {"type": "object"}, 0),
        ({"foo": datetime.datetime.now()}, 2, {"type": "object"}, 1),
        ({"foo": ["bar"]}, 20, {"type": "object"}, 0),
        ({"foo": ["bar"]}, 2, {"type": "object"}, 1),
        ({"foo": 1, "bar": {"Ref": "Parameter"}}, 20, {"type": "object"}, 0),
        ({"foo": 1, "bar": {"Ref": "Parameter"}}, 4, {"type": "object"}, 1),
        ({"foo": 1, "bar": {"Fn::Sub": "2"}}, 20, {"type": "object"}, 0),
        ({"foo": 1, "bar": {"Fn::Sub": "2"}}, 4, {"type": "object"}, 1),
        ({"foo": 1, "bar": {"Fn::Sub": ["2", {}]}}, 20, {"type": "object"}, 0),
        ({"foo": 1, "bar": {"Fn::Sub": ["2", {}]}}, 4, {"type": "object"}, 1),
    ],
)
def test_max_length(instance, mL, expected, rule, schema, validator):
    assert len(list(rule.maxLength(validator, mL, instance, schema))) == expected



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import defaultdict
from typing import Any

import cfnlint.helpers
from cfnlint._typing import Path, RuleMatches
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.template import Template


class RouteTableAssociation(CloudFormationLintRule):
    """Check only one route table association defined per subnet"""

    id = "E3022"
    shortdesc = "Resource SubnetRouteTableAssociation Properties"
    description = "Validate there is only one SubnetRouteTableAssociation per subnet"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-subnet-route-table-assoc.html"
    tags = ["resources", "ec2", "subnet", "route table"]

    def __init__(self) -> None:
        super().__init__()
        self.resource_values: dict[str, list[str]] = {}
        self.associated_resources: defaultdict[Any, list[str]] = defaultdict(list)

    def initialize(self, cfn: Template) -> None:
        self.resource_values = {}
        self.associated_resources = defaultdict(list)

    def get_values(self, subnetid, resource_condition, property_condition):
        """Get string literal(s) from value of SubnetId"""
        values = []
        if isinstance(subnetid, dict):
            if len(subnetid) == 1:
                for key, value in subnetid.items():
                    if key in cfnlint.helpers.CONDITION_FUNCTIONS:
                        if isinstance(value, list):
                            if len(value) == 3:
                                property_condition = value[0]
                                values.extend(
                                    self.get_values(
                                        value[1], resource_condition, property_condition
                                    )
                                )
                                values.extend(
                                    self.get_values(
                                        value[2], resource_condition, property_condition
                                    )
                                )
                    if key == "Ref":
                        values.extend(
                            self.get_values(
                                value, resource_condition, property_condition
                            )
                        )
                    if key == "Fn::GetAtt":
                        if isinstance(value[1], (str)):
                            sub_value = ".".join(value)
                            values.append(
                                (resource_condition, property_condition, sub_value)
                            )
        else:
            values.append((resource_condition, property_condition, subnetid))
        return values

    def check_values(self, subnetid, resource_condition, resource_name):
        """Check subnet value is not associated with other route tables"""
        property_condition = None
        values = self.get_values(subnetid, resource_condition, property_condition)
        self.resource_values[resource_name] = values
        for value in values:
            self.associated_resources[value].append(resource_name)

    def match(self, cfn: Template) -> RuleMatches:
        """Check SubnetRouteTableAssociation Resource Properties"""

        matches = []
        resources = cfn.get_resources(["AWS::EC2::SubnetRouteTableAssociation"])
        for resource_name, resource in resources.items():
            properties = resource.get("Properties")
            if properties:
                resource_condition = resource.get("Condition")
                subnetid = properties.get("SubnetId")
                self.check_values(subnetid, resource_condition, resource_name)
        for resource_name, resource_values in self.resource_values.items():
            for value in resource_values:
                bare_value = (None, None, value[2])
                other_resources: list[str] = []

                if len(self.associated_resources[value]) > 1:
                    for associated_name in self.associated_resources[value]:
                        if associated_name != resource_name:
                            other_resources.append(associated_name)

                if value != bare_value and self.associated_resources[bare_value]:
                    other_resources.extend(self.associated_resources[bare_value])

                if other_resources:
                    path: Path = ["Resources", resource_name, "Properties", "SubnetId"]
                    message = "SubnetId in {0} is also associated with {1}"
                    matches.append(
                        RuleMatch(
                            path,
                            message.format(resource_name, ", ".join(other_resources)),
                        )
                    )

        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import create_context_for_template
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.functions.GetAtt import GetAtt
from cfnlint.rules.functions.GetAz import GetAz
from cfnlint.rules.functions.Ref import Ref
from cfnlint.rules.functions.Sub import Sub
from cfnlint.rules.functions.SubResolved import SubResolved
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = Sub()
    rule.child_rules["W1031"] = SubResolved()
    yield rule


@pytest.fixture(scope="module")
def cfn():
    return Template(
        "",
        {
            "Parameters": {
                "MyParameter": {
                    "Type": "String",
                    "AllowedValues": [
                        "one",
                        "two",
                    ],
                }
            },
            "Resources": {
                "MyResource": {"Type": "AWS::S3::Bucket"},
                "MySimpleAd": {"Type": "AWS::DirectoryService::SimpleAD"},
                "MyPolicy": {"Type": "AWS::IAM::ManagedPolicy"},
            },
        },
        regions=["us-east-1"],
    )


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid Fn::Sub",
            {"Fn::Sub": "foo"},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Sub with an incorrect type",
            {"Fn::Sub": {"foo": "bar"}},
            {"type": "string"},
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'array', 'string'",
                    path=deque(["Fn::Sub"]),
                    schema_path=deque(["type"]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with an invalid output type",
            {"Fn::Sub": "foo"},
            {"type": "array"},
            [
                ValidationError(
                    "{'Fn::Sub': 'foo'} is not of type 'array'",
                    path=deque([]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Valid Fn::Sub with a valid Ref",
            {"Fn::Sub": "${AWS::Region}"},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with a valid Ref and a string escape",
            {"Fn::Sub": "${!Foo}=${AWS::Region}"},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with a valid Ref and a string escape with parameter",
            {"Fn::Sub": "${!foo}=${bar}"},
            {"type": "string"},
            [
                ValidationError(
                    (
                        "'bar' is not one of ["
                        "'MyParameter', 'MyResource', 'MySimpleAd', 'MyPolicy', "
                        "'AWS::AccountId', 'AWS::NoValue', 'AWS::NotificationARNs', "
                        "'AWS::Partition', 'AWS::Region', "
                        "'AWS::StackId', 'AWS::StackName', 'AWS::URLSuffix']"
                    ),
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a invalid Ref",
            {"Fn::Sub": "${foo}"},
            {"type": "string"},
            [
                ValidationError(
                    (
                        "'foo' is not one of ["
                        "'MyParameter', 'MyResource', 'MySimpleAd', 'MyPolicy', "
                        "'AWS::AccountId', 'AWS::NoValue', 'AWS::NotificationARNs', "
                        "'AWS::Partition', 'AWS::Region', "
                        "'AWS::StackId', 'AWS::StackName', 'AWS::URLSuffix']"
                    ),
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Valid Fn::Sub with a Ref to a sub parameter",
            {"Fn::Sub": ["${foo}", {"foo": "bar"}]},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with an escape character",
            {"Fn::Sub": "${!foo}"},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with an ref to resource",
            {"Fn::Sub": ["${MyResource}", {"MyResource": {"Ref": "MyResource"}}]},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Sub with an integer parameter",
            {"Fn::Sub": ["${Foo}", {"Foo": 1}]},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Sub with a too to many elements",
            {"Fn::Sub": ["${foo}", {"foo": "bar"}, {}]},
            {"type": "string"},
            [
                ValidationError(
                    "['${foo}', {'foo': 'bar'}, {}] is too long (2)",
                    path=deque(["Fn::Sub"]),
                    schema_path=deque(["maxItems"]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a bad object",
            {"Fn::Sub": ["${foo}", []]},
            {"type": "string"},
            [
                ValidationError(
                    "[] is not of type 'object'",
                    path=deque(["Fn::Sub", 1]),
                    schema_path=deque(["fn_items", "type"]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a bad object type",
            {
                "Fn::Sub": [
                    "${foo}",
                    {
                        "foo": [],
                    },
                ]
            },
            {"type": "string"},
            [
                ValidationError(
                    "[] is not of type 'string'",
                    path=deque(["Fn::Sub", 1, "foo"]),
                    schema_path=deque(
                        ["fn_items", "patternProperties", "[a-zA-Z0-9]+", "type"]
                    ),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a bad object type from fn",
            {
                "Fn::Sub": [
                    "${foo}",
                    {
                        "foo": {"Fn::GetAZs": ""},
                    },
                ]
            },
            {"type": "string"},
            [
                ValidationError(
                    "{'Fn::GetAZs': ''} is not of type 'string'",
                    path=deque(["Fn::Sub", 1, "foo"]),
                    schema_path=deque(
                        ["fn_items", "patternProperties", "[a-zA-Z0-9]+", "fn_getazs"]
                    ),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Valid Fn::Sub with a GetAtt",
            {"Fn::Sub": "${MyResource.Arn}"},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Sub with a GetAtt and a bad attribute",
            {"Fn::Sub": "${MyResource.Foo}"},
            {"type": "string"},
            [
                ValidationError(
                    (
                        "'Foo' is not one of ['Arn', "
                        "'DomainName', "
                        "'DualStackDomainName', "
                        "'RegionalDomainName', "
                        "'WebsiteURL'] in ['us-east-1']"
                    ),
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a GetAtt and a bad resource name",
            {"Fn::Sub": "${Foo.Bar}"},
            {"type": "string"},
            [
                ValidationError(
                    "'Foo' is not one of ['MyResource', 'MySimpleAd', 'MyPolicy']",
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a GetAtt to an array of attributes",
            {"Fn::Sub": "${MySimpleAd.DnsIpAddresses}"},
            {"type": "string"},
            [
                ValidationError(
                    ("'MySimpleAd.DnsIpAddresses' is not of type 'string'"),
                    instance="MySimpleAd.DnsIpAddresses",
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "Invalid Fn::Sub with a GetAtt to an integer",
            {"Fn::Sub": "${MyPolicy.AttachmentCount}"},
            {"type": "string"},
            [
                ValidationError(
                    ("'MyPolicy.AttachmentCount' is not of type 'string'"),
                    instance="MySimpleAd.DnsIpAddresses",
                    path=deque(["Fn::Sub"]),
                    schema_path=deque([]),
                    validator="fn_sub",
                ),
            ],
        ),
        (
            "One valid resolution",
            {"Fn::Sub": "${MyParameter}"},
            {"type": "string", "const": "two"},
            [],
        ),
        (
            "No valid resolution",
            {"Fn::Sub": "${MyParameter}"},
            {"type": "string", "const": "three"},
            [
                ValidationError(
                    ("'three' was expected when 'Fn::Sub' is resolved"),
                    path=deque(["Fn::Sub"]),
                    schema_path=deque(["const"]),
                    validator="fn_sub",
                    rule=SubResolved(),
                ),
                ValidationError(
                    ("'three' was expected when 'Fn::Sub' is resolved"),
                    path=deque(["Fn::Sub"]),
                    schema_path=deque(["const"]),
                    validator="fn_sub",
                    rule=SubResolved(),
                ),
            ],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, context, cfn):
    validator = CfnTemplateValidator({}).extend(
        validators={
            "ref": Ref().ref,
            "fn_getatt": GetAtt().fn_getatt,
            "fn_getazs": GetAz().fn_getazs,
        }
    )(context=context, cfn=cfn)
    errs = list(rule.fn_sub(validator, schema, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.resources.PreviousGenerationInstanceType import (
    PreviousGenerationInstanceType,  # pylint: disable=E0401
)


class TestPreviousGenerationInstanceType(BaseRuleTestCase):
    def setUp(self):
        super(TestPreviousGenerationInstanceType, self).setUp()
        self.collection.register(PreviousGenerationInstanceType())

    def test_file_positive(self):
        self.helper_file_positive()

    def test_file_negative(self):
        self.helper_file_negative(
            "test/fixtures/templates/bad/previous_generation_instances.yaml", 4
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path
from cfnlint.rules.formats.ImageId import ImageId


@pytest.fixture(scope="module")
def rule():
    rule = ImageId()
    yield rule


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "Valid image id",
            "ami-abcd1234",
            deque([]),
            True,
        ),
        (
            "Valid image id long",
            "ami-abcdefa1234567890",
            deque([]),
            True,
        ),
        (
            "Valid image id using ssm",
            "resolve:ssm:/aws/service/ecs/optimized-ami/amazon-linux-2023/recommended/image_id",
            deque(
                [
                    "Resources",
                    "AWS::EC2::LaunchTemplate",
                    "Properties",
                    "LaunchTemplateData",
                    "ImageId",
                ]
            ),
            True,
        ),
        (
            "Invalid image id using ssm",
            "resolve:ssm:/aws/service/ecs/optimized-ami/amazon-linux-2023/recommended/image_id",
            deque([]),
            False,
        ),
        (
            "Valid image id in exception",
            "ami-abcd1234",
            deque(
                [
                    "Resources",
                    "AWS::EC2::LaunchTemplate",
                    "Properties",
                    "LaunchTemplateData",
                    "ImageId",
                ]
            ),
            True,
        ),
        (
            "Valid type",
            [],
            deque([]),
            True,
        ),
        ("Invalid image ID", "ami-abc", deque([]), False),
    ],
)
def test_validate(name, instance, path, expected, rule, validator):
    validator = validator.evolve(
        context=validator.context.evolve(
            path=Path(
                path=deque([]),
                cfn_path=path,
            ),
        )
    )

    result = rule.format(validator, instance)
    assert result == expected, f"Test {name!r} got {result!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.metadata.Configuration import Configuration


@pytest.fixture(scope="module")
def rule():
    rule = Configuration()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid Metadata",
            {"Foo": "Bar"},
            [],
        ),
        (
            "Wrong type",
            [],
            [
                ValidationError(
                    ("[] is not of type 'object'"),
                    validator="type",
                    schema_path=deque(["type"]),
                    rule=Configuration(),
                )
            ],
        ),
        (
            "Invalid with a null",
            {"foo": None},
            [
                ValidationError(
                    (
                        "None is not of type "
                        "'string', 'integer', 'object', "
                        "'array', 'boolean'"
                    ),
                    validator="type",
                    schema_path=deque(["additionalProperties", "type"]),
                    path=deque(["foo"]),
                    rule=Configuration(),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from cfnlint.rules import CloudFormationLintRule


class FormatKeyword(CloudFormationLintRule):

    def __init__(self, format: str | None = None) -> None:
        super().__init__()
        self.format_keyword = format
        self.parent_rules = ["E1103"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.cloudfront.DistributionTargetOriginId import (
    DistributionTargetOriginId,
)


@pytest.fixture(scope="module")
def rule():
    rule = DistributionTargetOriginId()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            [],  # wrong type should return no issues
            [],
        ),
        (
            {
                "DefaultCacheBehavior": {
                    "TargetOriginId": "origin-id",
                },
                "Origins": [
                    {
                        "Id": "origin-id",
                    },
                ],
            },
            [],
        ),
        (
            {
                "DefaultCacheBehavior": {
                    "TargetOriginId": "origin-id",
                },
                "Origins": [
                    {
                        "Id": "foo",
                    },
                    {
                        "Id": "origin-id",
                    },
                    {
                        "Id": "bar",
                    },
                ],
            },
            [],
        ),
        (
            {
                "DefaultCacheBehavior": {
                    "TargetOriginId": "origin-id",
                },
                "Origins": [
                    {
                        "Id": "foo",
                    },
                    {
                        "Id": "bar",
                    },
                ],
            },
            [
                ValidationError(
                    ("'origin-id' is not one of ['foo', 'bar']"),
                    rule=DistributionTargetOriginId(),
                    path=deque([]),
                    validator="enum",
                    path_override=deque(["DefaultCacheBehavior", "TargetOriginId"]),
                )
            ],
        ),
        (
            {
                "DefaultCacheBehavior": {
                    "TargetOriginId": {"Ref": "MyParameter"},
                },
                "Origins": [
                    {
                        "Id": "foo",
                    },
                    {
                        "Id": "bar",
                    },
                ],
            },
            [],
        ),
        (
            {
                "DefaultCacheBehavior": {
                    "TargetOriginId": "origin-id",
                },
                "Origins": [
                    {
                        "Id": "foo",
                    },
                    {
                        "Id": {"Ref": "MyParameter"},
                    },
                    {
                        "Id": "bar",
                    },
                ],
            },
            [],
        ),
        (
            {
                "DefaultCacheBehavior": {
                    "TargetOriginId": "origin-id",
                },
                "Origins": [
                    {
                        "Id": "foo",
                    },
                    {
                        "Id": "bar",
                    },
                ],
                "OriginGroups": {
                    "Items": [
                        {
                            "Id": "group-1",
                        },
                        {
                            "Id": "group-2",
                        },
                    ]
                },
            },
            [
                ValidationError(
                    (
                        "'origin-id' is not one of "
                        "['foo', 'bar', 'group-1', 'group-2']"
                    ),
                    rule=DistributionTargetOriginId(),
                    path=deque([]),
                    validator="enum",
                    path_override=deque(["DefaultCacheBehavior", "TargetOriginId"]),
                )
            ],
        ),
        (
            {
                "DefaultCacheBehavior": {
                    "TargetOriginId": "origin-id",
                },
                "Origins": [
                    {
                        "Id": "foo",
                    },
                    {
                        "Id": "bar",
                    },
                ],
                "OriginGroups": {
                    "Items": [
                        {
                            "Id": "group-1",
                        },
                        {
                            "Id": "origin-id",
                        },
                    ]
                },
            },
            [],
        ),
        (
            {
                "DefaultCacheBehavior": {
                    "TargetOriginId": "origin-id",
                },
                "Origins": [
                    {
                        "Id": "foo",
                    },
                    {
                        "Id": "bar",
                    },
                ],
                "OriginGroups": {
                    "Items": [
                        {
                            "Id": "group-1",
                        },
                        {
                            "Id": {"Ref": "MyParameter"},
                        },
                    ]
                },
            },
            [],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging
from test.testlib.testcase import BaseTestCase
from unittest.mock import patch

from cfnlint import ConfigMixIn
from cfnlint.runner import Runner

LOGGER = logging.getLogger("cfnlint")


class TestCli(BaseTestCase):
    """Test CLI with config"""

    def tearDown(self):
        """Setup"""
        for handler in LOGGER.handlers:
            LOGGER.removeHandler(handler)

    @patch("cfnlint.maintenance.update_documentation")
    def test_update_documentation(self, mock_maintenance):
        config = ConfigMixIn(["--update-documentation"])

        runner = Runner(config)

        with self.assertRaises(SystemExit) as e:
            runner.cli()

        self.assertEqual(e.exception.code, 0)
        mock_maintenance.assert_called_once()

    @patch("cfnlint.maintenance.update_resource_specs")
    def test_update_specs(self, mock_maintenance):
        config = ConfigMixIn(["--update-specs"])

        runner = Runner(config)

        with self.assertRaises(SystemExit) as e:
            runner.cli()

        self.assertEqual(e.exception.code, 0)
        mock_maintenance.assert_called_once()

    @patch("cfnlint.maintenance.update_iam_policies")
    def test_update_iam_policies(self, mock_maintenance):
        config = ConfigMixIn(["--update-iam-policies"])

        runner = Runner(config)

        with self.assertRaises(SystemExit) as e:
            runner.cli()

        self.assertEqual(e.exception.code, 0)
        mock_maintenance.assert_called_once()

    def test_list_rules(self):
        config = ConfigMixIn(["--list-rules"])

        runner = Runner(config)

        with self.assertRaises(SystemExit) as e:
            runner.cli()

        self.assertEqual(e.exception.code, 0)

    @patch("argparse.ArgumentParser.print_help")
    @patch("sys.stdin.isatty")
    def test_print_help(self, mock_isatty, mock_print_help):
        config = ConfigMixIn([])

        runner = Runner(config)
        mock_isatty.return_value = True
        with self.assertRaises(SystemExit) as e:
            runner.cli()

        self.assertEqual(e.exception.code, 1)
        mock_print_help.assert_called_once()
        mock_isatty.assert_called_once()

    def test_bad_regions(self):
        config = ConfigMixIn(
            [
                "--regions",
                "us-north-5",
                "--template",
                "test/fixtures/templates/good/generic.yaml",
            ]
        )

        runner = Runner(config)

        with self.assertRaises(SystemExit) as e:
            runner.cli()

        self.assertEqual(e.exception.code, 32)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import namedtuple
from typing import Any, Tuple

from cfnlint.helpers import ToPy, ensure_list, is_types_compatible
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules import CloudFormationLintRule

SchemaDetails = namedtuple("SchemaDetails", ["module", "filename"])
all_types = ("array", "boolean", "integer", "number", "object", "string")
singular_types = ("boolean", "integer", "number", "string")


class BaseFn(CloudFormationLintRule):
    def __init__(
        self,
        name: str = "",
        types: Tuple[str, ...] | None = None,
        functions: Tuple[str, ...] | None = None,
        resolved_rule: str = "",
    ) -> None:
        super().__init__()
        self.fn = ToPy(name)
        self.types = types or tuple([])
        self.functions = functions or tuple([])
        self.resolved_rule = resolved_rule
        self.child_rules[self.resolved_rule] = None

    def key_value(self, instance: dict[str, Any]) -> Tuple[str, Any]:
        return list(instance.keys())[0], instance.get(self.fn.name)

    def fix_errors(self, errs: ValidationResult) -> ValidationResult:
        for err in errs:
            if err.validator not in ["ref"]:  # fix the list to all validators
                err.validator = self.fn.py
            yield err

    def schema(self, validator: Validator, instance: Any) -> dict[str, Any]:
        return {
            "type": ["string"],
        }

    def validator(self, validator: Validator) -> Validator:
        return validator.evolve(
            context=validator.context.evolve(
                functions=self.functions,
            ),
            function_filter=validator.function_filter.evolve(
                add_cfn_lint_keyword=False,
            ),
        )

    def resolve(
        self,
        validator: Validator,
        s: Any,
        instance: Any,
        schema: Any,
    ) -> ValidationResult:
        if not self.resolved_rule:
            return

        key, _ = self.key_value(instance)

        validator = validator.evolve(
            context=validator.context.evolve(
                strict_types=False,
            ),
        )
        all_errs = []
        for value, v, resolve_err in validator.resolve_value(instance):
            if resolve_err:
                yield resolve_err
                continue
            errs = list(
                self.fix_errors(
                    v.descend(
                        instance=value,
                        schema=s,
                        path=key,
                    )
                )
            )

            if not errs:
                return

            for err in errs:
                err.message = err.message.replace(f"{value!r}", f"{instance!r}")
                err.message = f"{err.message} when {self.fn.name!r} is resolved"
                all_errs.append(err)

        for err in all_errs:
            if self.child_rules[self.resolved_rule]:
                err.rule = self.child_rules[self.resolved_rule]
                yield err

    def _resolve_ref(self, validator, schema) -> Any:

        resolve = getattr(validator.resolver, "resolve", None)
        ref = schema["$ref"]
        if resolve is None:
            resolved = validator.resolver.resolving(ref)
        else:
            _, resolved = validator.resolver.resolve(ref)

        return resolved

    def resolve_type(self, validator, schema) -> list[str]:
        if "type" in schema:
            return ensure_list(schema["type"])  # type: ignore

        if "$ref" in schema:
            resolved = self._resolve_ref(validator, schema)

            return self.resolve_type(validator, resolved)

        return []

    def validate_fn_output_types(
        self, validator: Validator, s: Any, instance: Any
    ) -> ValidationResult:
        tS = self.resolve_type(validator, s)
        if tS:
            if not is_types_compatible(self.types, tS):
                reprs = ", ".join(repr(type) for type in tS)
                yield ValidationError(f"{instance!r} is not of type {reprs}")

    def validate(
        self,
        validator: Validator,
        s: Any,
        instance: Any,
        schema: Any,
    ) -> ValidationResult:
        # validate this function will return the correct type
        errs = list(
            self.fix_errors(self.validate_fn_output_types(validator, s, instance))
        )

        key, value = self.key_value(instance)
        errs.extend(
            list(
                self.fix_errors(
                    self.validator(validator).descend(
                        value,
                        self.schema(validator, instance),
                        path=key,
                    )
                )
            )
        )

        if errs:
            yield from iter(errs)
            return

        yield from self.resolve(validator, s, instance, schema)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase

from cfnlint.rules import CloudFormationLintRule


class TestCloudFormationRule(BaseTestCase):
    """Test CloudFormation Rule"""

    def test_base(self):
        """Test Base Rule"""

        class TestRule(CloudFormationLintRule):
            """Def Rule"""

            id = "E1000"
            shortdesc = "Test Rule"
            description = "Test Rule Description"
            source_url = "https://github.com/aws-cloudformation/cfn-lint/"
            tags = ["resources"]

        self.assertEqual(TestRule.id, "E1000")
        self.assertEqual(TestRule.shortdesc, "Test Rule")
        self.assertEqual(TestRule.description, "Test Rule Description")
        self.assertEqual(
            TestRule.source_url,
            "https://github.com/aws-cloudformation/cfn-lint/",
        )
        self.assertEqual(TestRule.tags, ["resources"])

    def test_config(self):
        """Test Configuration"""

        class TestRule(CloudFormationLintRule):
            """Def Rule"""

            id = "E1000"
            shortdesc = "Test Rule"
            description = "Test Rule"
            source_url = "https://github.com/aws-cloudformation/cfn-lint/"
            tags = ["resources"]

            def __init__(self):
                """Init"""
                super(TestRule, self).__init__()
                self.config_definition = {
                    "testBoolean": {"default": True, "type": "boolean"},
                    "testString": {"default": "default", "type": "string"},
                    "testInteger": {"default": 1, "type": "integer"},
                    "testListBoolean": {
                        "type": "list",
                        "itemtype": "boolean",
                        "default": [False],
                    },
                    "testListString": {
                        "type": "list",
                        "itemtype": "string",
                        "default": ["bar"],
                    },
                    "testListInteger": {
                        "type": "list",
                        "itemtype": "integer",
                        "default": [0],
                    },
                }
                self.configure()

            def get_config(self):
                """Get the Config"""
                return self.config

        rule = TestRule()
        config = rule.get_config()
        self.assertTrue(config.get("testBoolean"))
        self.assertEqual(config.get("testString"), "default")
        self.assertEqual(config.get("testInteger"), 1)
        self.assertListEqual(config.get("testListBoolean"), [False])
        self.assertListEqual(config.get("testListString"), ["bar"])
        self.assertListEqual(config.get("testListInteger"), [0])

        rule.configure(
            {
                "testBoolean": "false",
                "testString": "new",
                "testInteger": 2,
                "testListBoolean": ["true"],
                "testListString": ["foo"],
                "testListInteger": [1],
            }
        )

        self.assertFalse(config.get("testBoolean"))
        self.assertEqual(config.get("testString"), "new")
        self.assertEqual(config.get("testInteger"), 2)
        self.assertListEqual(config.get("testListBoolean"), [True])
        self.assertListEqual(config.get("testListString"), ["foo"])
        self.assertListEqual(config.get("testListInteger"), [1])

    def test_experimental(self):
        """Test Configuration"""

        class TestRule(CloudFormationLintRule):
            """Def Rule"""

            id = "E1000"
            shortdesc = "Test Rule"
            description = "Test Rule"
            source_url = "https://github.com/aws-cloudformation/cfn-lint/"
            tags = ["resources"]

            def __init__(self):
                """Init"""
                super(TestRule, self).__init__()
                self.config_definition = {
                    "experimental": {"default": True, "type": "boolean"},
                }
                self.configure()

            def get_config(self):
                """Get the Config"""
                return self.config

        rule = TestRule()
        rule.configure({}, True)
        config = rule.get_config()
        self.assertTrue(config.get("experimental"))

        rule.configure({}, False)
        config = rule.get_config()
        self.assertFalse(config.get("experimental"))

    def test_non_configured_experimental(self):
        class TestRule(CloudFormationLintRule):
            """Def Rule"""

            id = "E1000"
            shortdesc = "Test Rule"
            description = "Test Rule"
            source_url = "https://github.com/aws-cloudformation/cfn-lint/"
            tags = ["resources"]

        rule = TestRule()

        rule.configure({}, True)
        self.assertFalse(rule.config.get("experimental"))

    def test_none_equal(self):
        class TestRule(CloudFormationLintRule):
            """Def Rule"""

            id = "E1000"
            shortdesc = "Test Rule"
            description = "Test Rule"
            source_url = "https://github.com/aws-cloudformation/cfn-lint/"
            tags = ["resources"]

        rule = TestRule()

        self.assertNotEqual(rule, None)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.functions._BaseFn import BaseFn, all_types


class Length(BaseFn):
    """Check if Length values are correct"""

    id = "E1030"
    shortdesc = "Length validation of parameters"
    description = "Making sure Fn::Length is configured correctly"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-length.html"
    tags = ["functions", "length"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::Length",
            ("integer",),
            (
                "Ref",
                "Fn::FindInMap",
                "Fn::Split",
                "Fn::If",
                "Fn::GetAZs",
            ),
        )

    def schema(self, validator: Validator, instance: Any) -> dict[str, Any]:
        return {
            "type": ["array"],
            "fn_items": {
                "functions": [
                    "Fn::If",
                    "Fn::Base64",
                    "Fn::FindInMap",
                    "Fn::Join",
                    "Fn::Select",
                    "Fn::Split",
                    "Fn::Sub",
                    "Fn::ToJsonString",
                    "Fn::GetAZs",
                    "Ref",
                ],
                "schema": {
                    "type": all_types,
                },
            },
        }

    def fn_length(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        if not validator.context.transforms.has_language_extensions_transform():
            yield ValidationError(
                (
                    f"{self.fn.name} is not supported without "
                    "'AWS::LanguageExtensions' transform"
                ),
                validator=self.fn.py,
                rule=self,
            )
            return

        for err in super().validate(validator, s, instance, schema):
            err.rule = self
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Any, Iterator

LOGGER = logging.getLogger(__name__)


@dataclass(frozen=True)
class Mappings:
    """
    This class holds a mapping
    """

    maps: dict[str, Map] = field(init=True, default_factory=dict)
    is_transform: bool = field(init=True, default=False)

    @classmethod
    def create_from_dict(cls, instance: Any) -> Mappings:

        if not isinstance(instance, dict):
            return cls({})
        try:
            result = {}
            is_transform = False
            for k, v in instance.items():
                if k == "Fn::Transform":
                    is_transform = True
                else:
                    result[k] = Map.create_from_dict(v)
            return cls(result, is_transform)
        except (ValueError, AttributeError) as e:
            LOGGER.debug(e, exc_info=True)
            return cls({})


@dataclass(frozen=True)
class _MappingSecondaryKey:
    """
    This class holds a mapping value
    """

    keys: dict[str, list[Any] | str | int | float] = field(
        init=True, default_factory=dict
    )
    is_transform: bool = field(init=True, default=False)

    def value(self, secondary_key: str):
        if secondary_key not in self.keys:
            raise KeyError(secondary_key)
        return self.keys[secondary_key]

    @classmethod
    def create_from_dict(cls, instance: Any) -> _MappingSecondaryKey:
        if not isinstance(instance, dict):
            return cls({})
        is_transform = False
        keys = {}
        for k, v in instance.items():
            if k == "Fn::Transform":
                is_transform = True
            elif isinstance(v, (str, list, int, float)):
                keys[k] = v
            else:
                continue
        return cls(keys, is_transform)


@dataclass(frozen=True)
class Map:
    """
    This class holds a mapping
    """

    keys: dict[str, _MappingSecondaryKey] = field(init=True, default_factory=dict)
    is_transform: bool = field(init=True, default=False)

    def find_in_map(self, top_key: str, secondary_key: str) -> Iterator[Any]:
        if top_key not in self.keys:
            raise KeyError(top_key)
        yield self.keys[top_key].value(secondary_key)

    @classmethod
    def create_from_dict(cls, instance: Any) -> Map:
        if not isinstance(instance, dict):
            return cls({})
        is_transform = False
        keys = {}
        for k, v in instance.items():
            if k == "Fn::Transform":
                is_transform = True
            else:
                keys[k] = _MappingSecondaryKey.create_from_dict(v)
        return cls(keys, is_transform)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.helpers import FUNCTIONS, ensure_list
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class AvailabilityZone(CfnLintKeyword):
    """Check Availibility Zone parameter checks"""

    id = "W3010"
    shortdesc = "Availability zone properties should not be hardcoded"
    description = "Check if an Availability Zone property is hardcoded."
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["parameters", "availabilityzone"]

    def __init__(self):
        """Init"""
        super().__init__(
            keywords=[
                "Resources/AWS::AutoScaling::AutoScalingGroup/Properties/AvailabilityZones/*",
                "Resources/AWS::DAX::Cluster/Properties/AvailabilityZones/*",
                "Resources/AWS::DMS::ReplicationInstance/Properties/AvailabilityZone",
                "Resources/AWS::EC2::Host/Properties/AvailabilityZone",
                "Resources/AWS::EC2::Instance/Properties/AvailabilityZone",
                "Resources/AWS::EC2::LaunchTemplate/Properties/LaunchTemplateData/Placement/AvailabilityZone",
                "Resources/AWS::EC2::SpotFleet/Properties/SpotFleetRequestConfigData/LaunchSpecifications/*/Placement/AvailabilityZone",
                "Resources/AWS::EC2::SpotFleet/Properties/SpotFleetRequestConfigData/LaunchTemplateConfigs/*/Overrides/*/AvailabilityZone",
                "Resources/AWS::EC2::Subnet/Properties/AvailabilityZone",
                "Resources/AWS::EC2::Volume/Properties/AvailabilityZone",
                "Resources/AWS::ElasticLoadBalancing::LoadBalancer/Properties/AvailabilityZones/*",
                "Resources/AWS::ElasticLoadBalancingV2::TargetGroup/Properties/Targets/*/AvailabilityZone",
                "Resources/AWS::EMR::Cluster/Properties/Instances/Placement/AvailabilityZone",
                "Resources/AWS::Glue::Connection/Properties/ConnectionInput/PhysicalConnectionRequirements/AvailabilityZone",
                "Resources/AWS::OpsWorks::Instance/Properties/AvailabilityZone",
                "Resources/AWS::RDS::DBCluster/Properties/AvailabilityZones/*",
                "Resources/AWS::RDS::DBInstance/Properties/AvailabilityZone",
            ]
        )
        self.exceptions = ["all"]

    def validate(
        self, validator: Validator, keywords: Any, zones: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not isinstance(zones, (str, list)):
            return

        # Skip rule if CDK
        if validator.cfn.is_cdk_template():
            return

        zones = ensure_list(zones)

        for zone in zones:
            if not validator.is_type(zone, "string"):
                continue

            if zone in self.exceptions:
                continue

            if any(fn in validator.context.path.path for fn in FUNCTIONS):
                continue

            yield ValidationError(
                f"Avoid hardcoding availability zones {zone!r}",
                rule=self,
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.context import Context
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.jsonschema.CfnLint import CfnLint
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class _FooError(CfnLintKeyword):
    id = "EXXXX"
    shortdesc = "A rule"
    description = "A rule"

    def __init__(self) -> None:
        super().__init__(["Foo"])

    def validate(self, validator, keywords, instance, schema):
        yield ValidationError("No Foo")


class _BarError(CfnLintKeyword):
    id = "EYYYY"
    shortdesc = "A rule"
    description = "A rule"

    def __init__(self) -> None:
        super().__init__(["Bar"])

    def validate(self, validator, keywords, instance, schema):
        yield ValidationError("No Bar", rule=self)


class _SubError(CfnLintKeyword):
    def __init__(self) -> None:
        super().__init__(["Bar"])


@pytest.fixture(scope="module")
def rule():
    rule = CfnLint()
    rule.child_rules["EXXXX"] = _FooError()
    rule.child_rules["EYYYY"] = _BarError()
    rule.child_rules["EZZZZ"] = _SubError()
    yield rule


@pytest.mark.parametrize(
    "name,keywords,expected_errs",
    [
        (
            "When no error in exception add it",
            ["Foo"],
            [
                ValidationError(
                    message="No Foo",
                    rule=_FooError(),
                ),
            ],
        ),
        (
            "When no error in exception add it",
            ["Foo", "Bar"],
            [
                ValidationError(
                    message="No Foo",
                    rule=_FooError(),
                ),
                ValidationError(
                    message="No Bar",
                    rule=_BarError(),
                ),
            ],
        ),
        (
            "When no error in exception add it",
            ["FooBar"],
            [],
        ),
    ],
)
def test_cfn_schema(name, keywords, expected_errs, rule):
    context = Context(regions=["us-east-1"])
    validator = CfnTemplateValidator(schema={}, context=context)

    errs = list(rule.cfnLint(validator, keywords, True, {}))
    assert errs == expected_errs, f"{name} failed {errs} did not match {expected_errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import regex as re

from cfnlint.jsonschema import Validator
from cfnlint.rules.formats.FormatKeyword import FormatKeyword


class SecurityGroupId(FormatKeyword):
    id = "E1150"
    shortdesc = "Validate security group format"
    description = (
        "Security groups have to ref/gettatt to a security "
        "group or has the valid pattern"
    )
    tags = []
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/format_keyword.md#AWS::EC2::SecurityGroup.GroupId"

    def __init__(self):
        super().__init__(format="AWS::EC2::SecurityGroup.GroupId")

    def format(self, validator: Validator, instance: Any) -> bool:
        if not isinstance(instance, str):
            return True

        if re.match(r"^sg-([a-fA-F0-9]{8}|[a-fA-F0-9]{17})$", instance):
            return True

        return False



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.Split import Split


@pytest.fixture(scope="module")
def rule():
    rule = Split()
    yield rule


@pytest.fixture
def template():
    return {
        "Resources": {
            "MyResource": {
                "Type": "AWS::S3::Bucket",
            },
        },
    }


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid Fn::Split with array",
            {"Fn::Split": ["foo", "bar"]},
            {"type": "array"},
            [],
        ),
        (
            "Invalid Fn::Split with wrong output type",
            {"Fn::Split": ["foo", "bar"]},
            {"type": "string"},
            [
                ValidationError(
                    "{'Fn::Split': ['foo', 'bar']} is not of type 'string'",
                    path=deque([]),
                    schema_path=deque([]),
                    validator="fn_split",
                ),
            ],
        ),
        (
            "Invalid Fn::Split is NOT a array",
            {"Fn::Split": "foo"},
            {"type": "array"},
            [
                ValidationError(
                    "'foo' is not of type 'array'",
                    path=deque(["Fn::Split"]),
                    schema_path=deque(["type"]),
                    validator="fn_split",
                ),
            ],
        ),
        (
            "Invalid Fn::Split using a function for delimiter",
            {"Fn::Split": [{"foo": "bar"}, "bar"]},
            {"type": "array"},
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'string'",
                    path=deque(["Fn::Split", 0]),
                    schema_path=deque(["fn_items", "type"]),
                    validator="fn_split",
                ),
            ],
        ),
        (
            "Invalid Fn::Split using an invalid function",
            {"Fn::Split": ["-", {"foo": "bar"}]},
            {"type": "array"},
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'string'",
                    path=deque(["Fn::Split", 1]),
                    schema_path=deque(["fn_items", "type"]),
                    validator="fn_split",
                ),
            ],
        ),
        (
            "Invalid Fn::Split using an invalid CFN function",
            {"Fn::Split": ["-", {"Fn::Split": ["-", "bar"]}]},
            {"type": "array"},
            [
                ValidationError(
                    "{'Fn::Split': ['-', 'bar']} is not of type 'string'",
                    path=deque(["Fn::Split", 1]),
                    schema_path=deque(["fn_items", "type"]),
                    validator="fn_split",
                )
            ],
        ),
        (
            "Valid Fn::Split with a valid function",
            {"Fn::Split": ["foo", {"Fn::Sub": "bar"}]},
            {"type": "array"},
            [],
        ),
        (
            "Invalid Fn::Split with a dynamic reference",
            {"Fn::Split": ["foo", "{{resolve:ssm:Foo:1}}"]},
            {"type": "array"},
            [
                ValidationError(
                    "'Fn::Split' does not support dynamic references",
                    path=deque(["Fn::Split", 1]),
                    validator="fn_split",
                )
            ],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.fn_split(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.context import create_context_for_template
from cfnlint.context.conditions._conditions import Conditions
from cfnlint.context.conditions.exceptions import Unsatisfiable
from cfnlint.template import Template


def template():
    return {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Parameters": {
            "Environment": {
                "Type": "String",
                "Default": "dev",
                "AllowedValues": ["dev", "prod"],
            },
            "Name": {
                "Type": "String",
                "Default": "bi",
                "AllowedValues": ["bi", "ai"],
            },
            "AmiId": {
                "Type": "String",
                "Default": "",
            },
        },
        "Conditions": {
            "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
            "IsNotUsEast1": {"Fn::Not": [{"Condition": "IsUsEast1"}]},
            "IsUsWest2": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-west-2"]},
            "IsProd": {"Fn::Equals": [{"Ref": "Environment"}, "prod"]},
            "IsDev": {"Fn::Equals": [{"Ref": "Environment"}, "dev"]},
            "IsAmiProvided": {"Fn::Not": [{"Fn::Equals": [{"Ref": "AmiId"}, ""]}]},
            "IsAi": {"Fn::Equals": [{"Ref": "Name"}, "ai"]},
            "IsBi": {"Fn::Equals": [{"Ref": "Name"}, "bi"]},
            "IsUsEast1OrUsWest2": {
                "Fn::Or": [
                    {"Condition": "IsUsEast1"},
                    {"Condition": "IsUsWest2"},
                ]
            },
            "IsUsEast1AndProd": {
                "Fn::And": [
                    {"Condition": "IsUsEast1"},
                    {"Condition": "IsProd"},
                ]
            },
            "IsStaticTrue": {"Fn::Equals": [1, "1"]},
            "IsStaticFalse": {"Fn::Equals": [1, "2"]},
            "TwoObjects": {
                "Fn::Equals": [
                    {"Ref": "Environment"},
                    {"Ref": "AWS::Region"},
                ]
            },
            "TwoObjectsSameHash": {
                "Fn::Equals": [
                    {"Ref": "AWS::Region"},
                    {"Ref": "Environment"},
                ]
            },
            "BadAwsFn": {"Ref": "Foo"},
            "BadFn": {"Key": "Foo"},
            "BadNotList": {"Fn::And": "Foo"},
            "UseBadFn": {"Condition": "BadFn"},
            "UseBadCondition": {"Condition": []},
            "EqualsTooLong": {"Fn::Equals": ["a", "b", "c"]},
            "EqualsWrongTypes": {"Fn::Equals": ["a", []]},
            "NotWithTwoItems": {
                "Fn::Not": [
                    {"Condition": "IsAi"},
                    {"Condition": "IsBi"},
                ]
            },
        },
    }


def test_conditions():
    cfn = Template(None, template(), regions=["us-east-1"])
    context = create_context_for_template(cfn)

    assert len(context.conditions.conditions) == 22
    assert context.conditions.conditions["IsStaticTrue"].fn_equals.is_static is True
    assert context.conditions.conditions["IsStaticFalse"].fn_equals.is_static is False
    assert (
        context.conditions.conditions["TwoObjects"].fn_equals.hash
        == context.conditions.conditions["TwoObjectsSameHash"].fn_equals.hash
    )
    assert context.conditions.conditions["TwoObjects"].fn_equals.is_static is None

    for k in [
        "IsUsEast1",
        "IsNotUsEast1",
        "IsUsWest2",
        "IsUsEast1OrUsWest2",
        "IsUsEast1AndProd",
    ]:
        assert context.conditions.conditions[k].is_region is True

    for k in ["IsAmiProvided", "IsProd", "IsDev", "IsAi", "IsBi"]:
        assert context.conditions.conditions[k].is_region is False

    for k in [
        "BadAwsFn",
        "BadFn",
        "BadNotList",
        "UseBadCondition",
        "EqualsTooLong",
        "EqualsWrongTypes",
        "NotWithTwoItems",
    ]:
        assert context.conditions.conditions[k].fn_equals.left.instance is None
        assert context.conditions.conditions[k].fn_equals.right.instance is None
        assert context.conditions.conditions[k].fn_equals.is_static is None
        assert context.conditions.conditions[k].is_region is False

    assert (
        context.conditions.conditions["UseBadFn"].condition.fn_equals.left.instance
        is None
    )
    assert (
        context.conditions.conditions["UseBadFn"].condition.fn_equals.right.instance
        is None
    )
    assert (
        context.conditions.conditions["UseBadFn"].condition.fn_equals.is_static is None
    )


@pytest.mark.parametrize(
    "current_status,new_status,expected",
    [
        ({}, {}, {}),
        ({"IsUsEast1": True}, {}, {"IsUsEast1": True}),
        (
            {"IsUsEast1": True},
            {"IsUsEast1": False},
            Unsatisfiable(
                new_status={"IsUsEast1": False}, current_status={"IsUsEast1": True}
            ),
        ),
        (
            {"IsUsEast1": True},
            {"IsNotUsEast1": True},
            Unsatisfiable(
                new_status={"IsNotUsEast1": True},
                current_status={"IsUsEast1": True},
            ),
        ),
        (
            {"IsUsEast1": False, "IsUsWest2": False},
            {"IsUsEast1OrUsWest2": True},
            Unsatisfiable(
                new_status={"IsUsEast1OrUsWest2": True},
                current_status={"IsUsEast1": True},
            ),
        ),
        (
            {"IsDev": False},
            {"IsProd": False},
            Unsatisfiable(new_status={"IsProd": False}, current_status={"IsDev": True}),
        ),
        ({}, {"BadFn": True}, {"BadFn": True}),
        ({}, {"BadFn": False}, {"BadFn": False}),
    ],
)
def test_condition_status(current_status, new_status, expected):
    cfn = Template(None, template(), regions=["us-east-1"])
    context = create_context_for_template(cfn)

    context = context.evolve(conditions=context.conditions.evolve(current_status))

    if isinstance(expected, Exception):
        with pytest.raises(ValueError):
            context.conditions.evolve(new_status)
    else:
        context = context.evolve(conditions=context.conditions.evolve(new_status))
        assert context.conditions.status == expected


@pytest.mark.parametrize(
    "current_status,instance,expected",
    [
        ({}, {"Foo": "Bar"}, [({"Foo": "Bar"}, {})]),
        (
            {},
            {"Fn::If": ["IsUsEast1", {"Foo": "Foo"}, {"Bar": "Bar"}]},
            [
                ({"Foo": "Foo"}, {"IsUsEast1": True}),
                ({"Bar": "Bar"}, {"IsUsEast1": False}),
            ],
        ),
        (
            {
                "IsUsEast1": True,
            },
            {"Fn::If": ["IsUsEast1", {"Foo": "Foo"}, {"Bar": "Bar"}]},
            [
                ({"Foo": "Foo"}, {"IsUsEast1": True}),
            ],
        ),
        (
            {
                "IsUsEast1": False,
            },
            {"Fn::If": ["IsUsEast1", {"Foo": "Foo"}, {"Bar": "Bar"}]},
            [
                ({"Bar": "Bar"}, {"IsUsEast1": False}),
            ],
        ),
        (
            {},
            {"Ref": "AWS::NoValue"},
            [
                (None, {}),
            ],
        ),
        (
            {},
            [{"Foo": {"Fn::If": ["IsUsEast1", "Foo", "Bar"]}}],
            [
                ([{"Foo": {"Fn::If": ["IsUsEast1", "Foo", "Bar"]}}], {}),
            ],
        ),
        (
            {},
            [{"Fn::If": ["IsUsEast1", {"Foo": "Bar"}, {"Ref": "AWS::NoValue"}]}],
            [
                ([{"Foo": "Bar"}], {"IsUsEast1": True}),
                ([], {"IsUsEast1": False}),
            ],
        ),
        (
            {"IsUsEast1": True, "IsProd": True},
            {
                "A": {"Fn::If": ["IsUsEast1AndProd", 1, 2]},
                "B": {"Fn::If": ["IsAi", 10, 11]},
            },
            [
                (
                    {"A": 1, "B": 10},
                    {
                        "IsUsEast1": True,
                        "IsProd": True,
                        "IsUsEast1AndProd": True,
                        "IsAi": True,
                    },
                ),
                (
                    {"A": 1, "B": 11},
                    {
                        "IsUsEast1": True,
                        "IsProd": True,
                        "IsUsEast1AndProd": True,
                        "IsAi": False,
                    },
                ),
            ],
        ),
        (
            {},
            {
                "A": {"Fn::If": ["IsUsEast1AndProd", 1]},
            },
            [
                (
                    {"A": {"Fn::If": ["IsUsEast1AndProd", 1]}},
                    {},
                ),
            ],
        ),
    ],
)
def test_evolve_from_instance(current_status, instance, expected):
    cfn = Template(None, template(), regions=["us-east-1"])
    context = create_context_for_template(cfn)

    context = context.evolve(
        conditions=context.conditions.evolve(current_status),
        functions=["Fn::If", "Ref"],
    )

    results = list(context.conditions.evolve_from_instance(instance, context))
    assert len(results) == len(expected)
    for result, expected_result in zip(results, expected):
        assert result[0] == expected_result[0]
        assert result[1].status == expected_result[1]


def test_condition_failures():
    with pytest.raises(ValueError):
        Conditions.create_from_instance([], {})



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.metadata.CfnLint import CfnLint


@pytest.fixture(scope="module")
def rule():
    rule = CfnLint()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid cfn-lint config",
            {"config": {"ignore_checks": ["E3002"]}},
            [],
        ),
        (
            "Extra properties",
            {"config": {"ignore_bad_key": ["E3002"]}},
            [
                ValidationError(
                    (
                        "Additional properties are not allowed "
                        "('ignore_bad_key' was unexpected)"
                    ),
                    validator="additionalProperties",
                    schema_path=deque(["properties", "config", "additionalProperties"]),
                    rule=CfnLint(),
                    path=deque(["config", "ignore_bad_key"]),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules.jsonschema.PropertyNames import PropertyNames as ParentPropertyNames


class MaxLength(ParentPropertyNames):
    """Check if maximum Resource name size limit is exceeded"""

    id = "E3011"
    shortdesc = "Check property names in Resources"
    description = "Validate property names are property configured in Resources"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cloudformation-limits.html"
    tags = ["resources", "limits"]

    def __init__(self) -> None:
        super().__init__(
            approaching_limit_rule="I3012",
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema.validators import CfnTemplateValidator
from cfnlint.rules.resources.iam.RoleArnPattern import RoleArnPattern


def _message_errors(name, arn, errors, **kwargs):
    validator = CfnTemplateValidator().evolve(**kwargs)

    i_errors = list(RoleArnPattern().validate(validator, {}, arn, {}))

    assert len(errors) == len(i_errors), (
        f"{name}: Expected exactly {len(errors)} error, "
        f"found {i_errors!r}, need {errors!r}"
    )

    err_messages = [err.message for err in i_errors]
    for err in errors:
        assert err in err_messages, f"{name}: expected {err} to be in {i_errors!r}"


@pytest.mark.parametrize(
    "name,arn,errors",
    [
        (
            "Invalid Arn",
            "test",
            [
                (
                    "'test' does not match '^arn:(aws[a-zA-Z-]*)?:"
                    "iam::\\\\d{12}:role/[a-zA-Z_0-9+=,.@\\\\-_/]+$'"
                )
            ],
        ),
        (
            "Valid but wrong type",
            {},
            [],
        ),
        (
            "Valid Arn",
            "arn:aws:iam::123456789012:role/test",
            [],
        ),
        (
            "Valid Arn for aws-us-gov",
            "arn:aws-us-gov:iam::123456789012:role/test",
            [],
        ),
    ],
)
def test_iam_role_arn(name, arn, errors):
    _message_errors(name, arn, errors)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.rules.resources.rds.DbInstanceDbInstanceClassEnum import (
    DbInstanceDbInstanceClassEnum,
)


@pytest.fixture(scope="module")
def rule():
    rule = DbInstanceDbInstanceClassEnum()
    yield rule


@pytest.mark.parametrize(
    "name,instance,regions,err_count",
    [
        (
            "Valid version",
            {"DBInstanceClass": "db.x2iedn.8xlarge", "Engine": "mysql"},
            ["us-east-1"],
            0,
        ),
        (
            "Valid but wrong type",
            [],
            ["us-east-1"],
            0,
        ),
        (
            "Valid but wrong type for Engine",
            {"Engine": []},
            ["us-east-1"],
            0,
        ),
        (
            "No error ref value",
            {
                "DBInstanceClass": "notanappropriatetype",
                "Engine": "sqlserver-se",
                "LicenseModel": {"Ref": "License"},
            },
            ["us-east-1"],
            0,
        ),
        (
            "Bad instance type with no license",
            {"DBInstanceClass": "db.m4.xlarge", "Engine": "aurora-mysql"},
            ["us-east-1"],
            1,
        ),
        (
            "Bad value on bring your own license",
            {
                "DBInstanceClass": "db.x2e.xlarge",
                "Engine": "oracle-se2",
                "LicenseModel": "bring-your-own-license",
            },
            ["us-east-1"],
            1,
        ),
        (
            "Bad value on bring your own license",
            {
                "DBInstanceClass": "db.x2e.xlarge",
                "Engine": "oracle-se2",
                "LicenseModel": "bring-your-own-license",
            },
            ["us-east-1"],
            1,
        ),
        (
            "Bad value on license included",
            {
                "DBInstanceClass": "db.x1e.xlarge",
                "Engine": "oracle-se2",
                "LicenseModel": "license-included",
            },
            ["us-east-1"],
            1,
        ),
        (
            "Bad value on license included and uppser case Engine",
            {
                "DBInstanceClass": "db.x1e.xlarge",
                "Engine": "Oracle-SE2",
                "LicenseModel": "license-included",
            },
            ["us-east-1"],
            1,
        ),
        (
            "Instance type invalid in one region",
            {"DBInstanceClass": "db.x2iedn.8xlarge", "Engine": "mysql"},
            ["us-east-1", "ca-west-1"],
            1,
        ),
    ],
)
def test_validate(name, instance, regions, err_count, rule, validator):
    validator = validator.evolve(context=validator.context.evolve(regions=regions))
    errors = list(rule.validate(validator, "", instance, {}))

    # we use error counts in this one as the instance types are
    # always changing so we aren't going to hold ourselves up by that
    assert len(errors) == err_count, f"Test {name!r} got {errors!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import unittest

from cfnlint.jsonschema._utils import equal, uniq, uniq_keys


class TestUtils(unittest.TestCase):
    def test_equal(self):
        self.assertTrue(equal("foo", "foo"))
        self.assertFalse(equal("foo", "bar"))
        self.assertTrue(equal(["foo"], ["foo"]))
        self.assertFalse(equal(["foo"], ["bar"]))
        self.assertFalse(equal(["foo"], ["foo", "bar"]))
        self.assertTrue(equal({"foo": "bar"}, {"foo": "bar"}))
        self.assertFalse(equal({"foo": "bar"}, {"bar": "foo"}))
        self.assertFalse(equal({"foo": "bar"}, {"foo": "bar", "bar": "foo"}))
        self.assertTrue(equal(True, True))
        self.assertFalse(equal(True, False))
        self.assertTrue(equal(1, 1))
        self.assertFalse(equal(1, 2))

        self.assertTrue(equal(1, "1"))
        self.assertFalse(equal({"foo": "bar"}, "1"))

    def test_uniq(self):
        self.assertTrue(uniq(["foo", "bar"]))
        self.assertFalse(uniq(["foo", "foo"]))
        self.assertFalse(uniq([{"foo": "bar"}, {"foo": "bar"}]))
        self.assertTrue(uniq([{"foo": "bar"}, ["foo", "bar"]]))
        self.assertTrue(uniq([{"foo": "bar"}, {"bar": "foo"}]))

        self.assertFalse(uniq(["1", 1]))

    def test_uniq_keys(self):
        # Bad structure
        self.assertTrue(uniq_keys(["foo", "bar"], ["foo"]))

        # Unknown keys
        self.assertTrue(uniq_keys([{"foo": "foo"}, {"bar": "bar"}], ["key"]))

        # Valid and invalid configuration
        self.assertTrue(uniq_keys([{"foo": "foo"}, {"foo": "bar"}], ["foo"]))
        self.assertFalse(uniq_keys([{"foo": "foo"}, {"foo": "foo"}], ["foo"]))
        self.assertFalse(
            uniq_keys([{"foo": "foo"}, {"bar": "bar"}, {"foo": "foo"}], ["foo"])
        )
